# -*- coding: utf-8 -*-

# Copyright (c) 2019, Brandon Nielsen
#
# This file is part of Railyard.
#
# Railyard is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Railyard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Railyard.  If not, see <https://www.gnu.org/licenses/>.

import argparse
import os
import socket
import subprocess

import yaml

from railyard.app import RailyardApplication

__version__ = "0.0.3"


def _parse_routes(dispatch_yaml):
    routes = {}

    dispatch_dict = yaml.safe_load(dispatch_yaml)

    for route in dispatch_dict["dispatch"]:
        routes[route["url"]] = route["service"]

    return routes


def _get_port():
    # Bind a socket on the first open port provided by the OS
    net_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    net_socket.bind(("", 0))

    port = net_socket.getsockname()[1]

    net_socket.close()

    return port


def _parse_app(app_file_path):
    routes = {"*": "default"}
    services = {}  # Key: service name, value: service yaml as dict
    working_directories = {}  # Key: service name, value: working directory for service

    with open(app_file_path, "r") as app_yaml:
        services["default"] = yaml.safe_load(app_yaml)
        working_directories["default"] = "."

    return routes, services, working_directories


def _parse_dispatch(dispatch_file_path):
    routes = {}  # Key: route, value: service name
    services = {}  # Key: service name, value: service yaml as dict
    working_directories = {}  # Key: service name, value: working directory for service

    # Read the dispatch file
    with open(dispatch_file_path, "r") as dispatch_yaml:
        routes = _parse_routes(dispatch_yaml)

    for route in routes:
        service = routes[route]

        if service == "default":
            with open("app.yaml", "r") as app_yaml:
                services["default"] = yaml.safe_load(app_yaml)
                working_directories["default"] = "."
        else:
            # Check for '$SERVICE.yaml' first, than '$SERVICE/$SERVICE.yaml'
            service_filename = service + ".yaml"
            service_file_path = os.path.join(service, service_filename)

            if os.path.isfile(service_filename) is True:
                with open(service_filename, "r") as app_yaml:
                    services[service] = yaml.safe_load(app_yaml)
                    working_directories[service] = "."
            elif os.path.isfile(service_file_path) is True:
                with open(service_file_path, "r") as app_yaml:
                    services[service] = yaml.safe_load(app_yaml)
                    working_directories[service] = service
            else:
                raise RuntimeError("Unable to find {}".format(service_filename))

    return routes, services, working_directories


def _launch_services(services, working_directories):
    ports = {}  # Key: service name, value: port for service

    for service in services:
        service_dict = services[service]
        port = _get_port()

        # Determine how to launch the service
        if "entrypoint" in service_dict:
            popen_args = service_dict["entrypoint"]
        else:
            popen_args = "gunicorn -b :$PORT main:app"

        # Set environment variables
        popen_env = os.environ.copy()
        popen_env["PORT"] = str(port)

        # Launch the service
        subprocess.Popen(
            popen_args, shell=True, cwd=working_directories[service], env=popen_env
        )

        ports[service] = port

    return ports


def main():
    parser = argparse.ArgumentParser(
        description="Runner for local development of standard environment App Engine applications."
    )

    parser.add_argument(
        "input",
        type=str,
        choices=["app.yaml", "dispatch.yaml"],
        help="the app.yaml or dispatch.yaml file to run",
    )

    args = parser.parse_args()

    if args.input == "app.yaml":
        routes, services, working_directories = _parse_app(args.input)
    elif args.input == "dispatch.yaml":
        routes, services, working_directories = _parse_dispatch(args.input)
    else:
        raise RuntimeError("Unrecognized input file {}".format(args.input))

    ports = _launch_services(services, working_directories)

    # Spawn the proxy
    RailyardApplication(
        routes, services, working_directories, ports, {"bind": ":8000"}
    ).run()



# -*- coding: utf-8 -*-

# Copyright (c) 2019, Brandon Nielsen
#
# This file is part of Railyard.
#
# Railyard is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Railyard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Railyard.  If not, see <https://www.gnu.org/licenses/>.

import argparse
import os
import socket
import subprocess
import unittest
from collections import namedtuple
from unittest.mock import Mock, call, mock_open, patch

import railyard
from railyard import (
    _get_port,
    _launch_services,
    _parse_app,
    _parse_dispatch,
    _parse_routes,
    main,
)


class TestInitFunctions(unittest.TestCase):
    def test_parse_routes(self):
        with open(
            os.path.join(os.path.dirname(__file__), "test_dispatch1.yaml"), "r"
        ) as yaml_contents:
            routes = _parse_routes(yaml_contents)

        self.assertEqual(
            routes, {"*/mobile/*": "mobile-frontend", "*/work/*": "static-backend"}
        )

    def test_get_port(self):
        mock_socket = Mock()
        mock_socket.getsockname.return_value = ["dummy", 1234]

        with patch.object(socket, "socket") as mock_socket_constructor:
            mock_socket_constructor.return_value = mock_socket

            result = _get_port()

        self.assertEqual(result, 1234)

    def test_parse_app(self):
        app_file_path = os.path.join(os.path.dirname(__file__), "test_app1.yaml")

        with open(app_file_path, "r") as yaml_contents:
            with patch("builtins.open") as mock_open_yaml:
                mock_open_yaml.side_effect = mock_open(read_data=yaml_contents.read())

                results = _parse_app(app_file_path)

        mock_open_yaml.assert_called_once_with(app_file_path, "r")

        self.assertEqual(
            results,
            (
                {"*": "default"},
                {
                    "default": {
                        "env_variables": {"BUCKET_NAME": "example-gcs-bucket"},
                        "handlers": [
                            {"static_dir": "static/images", "url": "/images"},
                            {
                                "redirect_http_response_code": 301,
                                "script": "auto",
                                "secure": "always",
                                "url": "/.*",
                            },
                        ],
                        "instance_class": "F2",
                        "runtime": "python37",
                    }
                },
                {"default": "."},
            ),
        )

    def test_parse_dispatch(self):
        # Test with default service
        dispatch_file_path = os.path.join(
            os.path.dirname(__file__), "test_dispatch2.yaml"
        )
        app1_file_path = os.path.join(os.path.dirname(__file__), "test_app1.yaml")

        with open(dispatch_file_path, "r") as yaml_contents:
            dispatch_file_contents = yaml_contents.read()

        with open(app1_file_path, "r") as yaml_contents:
            app1_file_contents = yaml_contents.read()

        with patch("builtins.open") as mock_open_yaml:
            mock_open_yaml.side_effect = (
                mock_open(read_data=dispatch_file_contents).return_value,
                mock_open(read_data=app1_file_contents).return_value,
            )

            results = _parse_dispatch(dispatch_file_path)

        self.assertEqual(
            mock_open_yaml.mock_calls,
            [call(dispatch_file_path, "r"), call("app.yaml", "r")],
        )

        self.assertEqual(
            results,
            (
                {"*": "default"},
                {
                    "default": {
                        "env_variables": {"BUCKET_NAME": "example-gcs-bucket"},
                        "handlers": [
                            {"static_dir": "static/images", "url": "/images"},
                            {
                                "redirect_http_response_code": 301,
                                "script": "auto",
                                "secure": "always",
                                "url": "/.*",
                            },
                        ],
                        "instance_class": "F2",
                        "runtime": "python37",
                    }
                },
                {"default": "."},
            ),
        )

        # Test with app files in dispatch folder
        dispatch_file_path = os.path.join(
            os.path.dirname(__file__), "test_dispatch3.yaml"
        )
        app2_file_path = os.path.join(os.path.dirname(__file__), "test_app2.yaml")
        app3_file_path = os.path.join(os.path.dirname(__file__), "test_app3.yaml")

        with open(dispatch_file_path, "r") as yaml_contents:
            dispatch_file_contents = yaml_contents.read()

        with open(app2_file_path, "r") as yaml_contents:
            app2_file_contents = yaml_contents.read()

        with open(app3_file_path, "r") as yaml_contents:
            app3_file_contents = yaml_contents.read()

        with patch("builtins.open") as mock_open_yaml:
            with patch.object(os.path, "isfile") as mock_os_path_isfile:
                mock_open_yaml.side_effect = (
                    mock_open(read_data=dispatch_file_contents).return_value,
                    mock_open(read_data=app2_file_contents).return_value,
                    mock_open(read_data=app3_file_contents).return_value,
                )
                mock_os_path_isfile.side_effect = (True, True)

                results = _parse_dispatch(dispatch_file_path)

        self.assertEqual(
            mock_open_yaml.mock_calls,
            [
                call(dispatch_file_path, "r"),
                call("test_app2.yaml", "r"),
                call("test_app3.yaml", "r"),
            ],
        )
        self.assertEqual(
            mock_os_path_isfile.mock_calls,
            [call("test_app2.yaml"), call("test_app3.yaml")],
        )

        self.assertEqual(
            results,
            (
                {"*/app2/*": "test_app2", "*/app3/*": "test_app3"},
                {
                    "test_app2": {
                        "entrypoint": "gunicorn -b :$PORT main:app",
                        "handlers": [{"static_dir": "static/", "url": "/app2/static"}],
                        "runtime": "python37",
                    },
                    "test_app3": {
                        "entrypoint": "gunicorn -b :$PORT main:app",
                        "handlers": [
                            {
                                "static_files": "static/\\1",
                                "url": "/app3/(.*\\.(gif|png|jpg))$",
                            }
                        ],
                        "runtime": "python37",
                    },
                },
                {"test_app2": ".", "test_app3": "."},
            ),
        )

        # Test with app files in separate folder
        with patch("builtins.open") as mock_open_yaml:
            with patch.object(os.path, "isfile") as mock_os_path_isfile:
                mock_open_yaml.side_effect = (
                    mock_open(read_data=dispatch_file_contents).return_value,
                    mock_open(read_data=app2_file_contents).return_value,
                    mock_open(read_data=app3_file_contents).return_value,
                )
                mock_os_path_isfile.side_effect = (False, True, False, True)

                results = _parse_dispatch(dispatch_file_path)

        self.assertEqual(
            mock_open_yaml.mock_calls,
            [
                call(dispatch_file_path, "r"),
                call(os.path.join("test_app2", "test_app2.yaml"), "r"),
                call(os.path.join("test_app3", "test_app3.yaml"), "r"),
            ],
        )
        self.assertEqual(
            mock_os_path_isfile.mock_calls,
            [
                call("test_app2.yaml"),
                call(os.path.join("test_app2", "test_app2.yaml")),
                call("test_app3.yaml"),
                call(os.path.join("test_app3", "test_app3.yaml")),
            ],
        )

        self.assertEqual(
            results,
            (
                {"*/app2/*": "test_app2", "*/app3/*": "test_app3"},
                {
                    "test_app2": {
                        "entrypoint": "gunicorn -b :$PORT main:app",
                        "handlers": [{"static_dir": "static/", "url": "/app2/static"}],
                        "runtime": "python37",
                    },
                    "test_app3": {
                        "entrypoint": "gunicorn -b :$PORT main:app",
                        "handlers": [
                            {
                                "static_files": "static/\\1",
                                "url": "/app3/(.*\\.(gif|png|jpg))$",
                            }
                        ],
                        "runtime": "python37",
                    },
                },
                {"test_app2": "test_app2", "test_app3": "test_app3"},
            ),
        )

        # Test missing dispatch
        with patch("builtins.open") as mock_open_yaml:
            with patch.object(os.path, "isfile") as mock_os_path_isfile:
                mock_open_yaml.side_effect = (
                    mock_open(read_data=dispatch_file_contents).return_value,
                    mock_open(read_data=app2_file_contents).return_value,
                    mock_open(read_data=app3_file_contents).return_value,
                )
                mock_os_path_isfile.side_effect = (False, False)

                with self.assertRaises(RuntimeError):
                    _parse_dispatch(dispatch_file_path)

        self.assertEqual(
            mock_open_yaml.mock_calls, [call(dispatch_file_path, "r"),],
        )
        self.assertEqual(
            mock_os_path_isfile.mock_calls,
            [
                call("test_app2.yaml"),
                call(os.path.join("test_app2", "test_app2.yaml")),
            ],
        )

    def test_launch_services(self):
        test_services = {"service1": {}, "service2": {"entrypoint": "service2 $PORT"}}
        test_working_directories = {"service1": ".", "service2": "service2"}

        with patch.object(railyard, "_get_port") as mock_get_port:
            with patch.object(os.environ, "copy") as mock_environ_copy:
                with patch.object(subprocess, "Popen") as mock_popen:
                    mock_get_port.side_effect = (1111, 2222)
                    mock_environ_copy.side_effect = (
                        {"test": "environment"},
                        {"test": "environment"},
                    )

                    result_ports = _launch_services(
                        test_services, test_working_directories
                    )

        self.assertEqual(
            mock_popen.mock_calls,
            [
                call(
                    "gunicorn -b :$PORT main:app",
                    shell=True,
                    cwd=".",
                    env={"test": "environment", "PORT": "1111"},
                ),
                call(
                    "service2 $PORT",
                    shell=True,
                    cwd="service2",
                    env={"test": "environment", "PORT": "2222"},
                ),
            ],
        )

        self.assertEqual(result_ports, {"service1": 1111, "service2": 2222})

    def test_main(self):
        # Test with app.yaml
        mock_argument_parser = Mock()
        mock_argument_parser.parse_args.return_value = _dict_to_argument_parser_result(
            {"input": "app.yaml"}
        )

        mock_railyard_application = Mock()

        with patch.object(
            argparse, "ArgumentParser"
        ) as mock_argument_parser_constructor:
            with patch.object(railyard, "_parse_app") as mock_parse_app:
                with patch.object(railyard, "_launch_services") as mock_launch_services:
                    with patch.object(
                        railyard, "RailyardApplication"
                    ) as mock_railyard_application_constructor:
                        mock_argument_parser_constructor.return_value = (
                            mock_argument_parser
                        )
                        mock_parse_app.return_value = (
                            {"routes": "routes data"},
                            {"services": "services data"},
                            {"working_directories": "working directories data"},
                        )
                        mock_launch_services.return_value = {"ports": "ports data"}
                        mock_railyard_application_constructor.return_value = (
                            mock_railyard_application
                        )

                        main()

        # Make sure the arg parser is correctly constructed
        self.assertEqual(
            mock_argument_parser.add_argument.mock_calls,
            [
                call(
                    "input",
                    type=str,
                    choices=["app.yaml", "dispatch.yaml"],
                    help="the app.yaml or dispatch.yaml file to run",
                )
            ],
        )

        mock_parse_app.assert_called_once_with("app.yaml")
        mock_launch_services.assert_called_once_with(
            {"services": "services data"},
            {"working_directories": "working directories data"},
        )
        mock_railyard_application_constructor.assert_called_once_with(
            {"routes": "routes data"},
            {"services": "services data"},
            {"working_directories": "working directories data"},
            {"ports": "ports data"},
            {"bind": ":8000"},
        )
        mock_railyard_application.run.assert_called_once_with()

        # Test with dispatch.yaml
        mock_argument_parser = Mock()
        mock_argument_parser.parse_args.return_value = _dict_to_argument_parser_result(
            {"input": "dispatch.yaml"}
        )

        mock_railyard_application = Mock()

        with patch.object(
            argparse, "ArgumentParser"
        ) as mock_argument_parser_constructor:
            with patch.object(railyard, "_parse_dispatch") as mock_parse_dispatch:
                with patch.object(railyard, "_launch_services") as mock_launch_services:
                    with patch.object(
                        railyard, "RailyardApplication"
                    ) as mock_railyard_application_constructor:
                        mock_argument_parser_constructor.return_value = (
                            mock_argument_parser
                        )
                        mock_parse_dispatch.return_value = (
                            {"routes": "routes data"},
                            {"services": "services data"},
                            {"working_directories": "working directories data"},
                        )
                        mock_launch_services.return_value = {"ports": "ports data"}
                        mock_railyard_application_constructor.return_value = (
                            mock_railyard_application
                        )

                        main()

        mock_parse_dispatch.assert_called_once_with("dispatch.yaml")
        mock_launch_services.assert_called_once_with(
            {"services": "services data"},
            {"working_directories": "working directories data"},
        )
        mock_railyard_application_constructor.assert_called_once_with(
            {"routes": "routes data"},
            {"services": "services data"},
            {"working_directories": "working directories data"},
            {"ports": "ports data"},
            {"bind": ":8000"},
        )
        mock_railyard_application.run.assert_called_once_with()

        # Test with invalid input
        mock_argument_parser = Mock()
        mock_argument_parser.parse_args.return_value = _dict_to_argument_parser_result(
            {"input": "bad.yaml"}
        )

        with patch.object(
            argparse, "ArgumentParser"
        ) as mock_argument_parser_constructor:
            mock_argument_parser_constructor.return_value = mock_argument_parser

            with self.assertRaises(RuntimeError):
                main()


def _dict_to_argument_parser_result(args_dict):
    return namedtuple("Args", args_dict.keys())(*args_dict.values())



# -*- coding: utf-8 -*-

# Copyright (c) 2019, Brandon Nielsen
#
# This file is part of Railyard.
#
# Railyard is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Railyard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Railyard.  If not, see <https://www.gnu.org/licenses/>.

import fnmatch
import mimetypes
import mmap
import os
import re
import urllib
import urllib.request

import gunicorn.app.base


class RailyardApplication(gunicorn.app.base.BaseApplication):
    def __init__(self, routes, services, working_directories, ports, options=None):
        # http://docs.gunicorn.org/en/stable/custom.html
        self.options = options or {}

        self._port_mappings = {}  # key: route, value: port
        self._url_handlers = (
            {}
        )  # key: route, value: dict (key: url, value: handler dict)
        self._working_directory_mappings = {}  # key: route, value: working directory

        # Build the port mapping and working directory dicts
        for route in routes:
            service_name = routes[route]

            self._port_mappings[route] = ports[service_name]
            self._working_directory_mappings[route] = working_directories[service_name]

            # Build the path handlers dict
            handlers_dict = {}  # key: url, value: handler dict

            if "handlers" in services[service_name]:
                for handler in services[service_name]["handlers"]:
                    handlers_dict[handler["url"]] = handler

            self._url_handlers[route] = handlers_dict

        super(RailyardApplication, self).__init__()

    def load_config(self):
        config = dict(
            [
                (key, value)
                for key, value in self.options.items()
                if key in self.cfg.settings and value is not None
            ]
        )
        for key, value in config.items():
            self.cfg.set(key.lower(), value)

    def load(self):
        return self

    def __call__(self, environ, start_response):
        matching_route = None
        matching_url = None

        request_path = environ.get("PATH_INFO", "")

        for route in self._url_handlers:
            if _match_route(_build_request_url(environ), route):
                matching_route = route
                break

        # Fail early
        if matching_route is None:
            return _not_found(environ, start_response)

        for url in self._url_handlers[matching_route]:
            if _match_url(request_path, url):
                matching_url = url
                break

        if matching_url is not None:
            handler_dict = self._url_handlers[matching_route][matching_url]

            # Handle static files specially
            static_file_path = None

            if "static_dir" in handler_dict:
                static_file_path = _build_file_path(
                    request_path,
                    matching_url,
                    static_dir=handler_dict["static_dir"],
                    working_directory=self._working_directory_mappings[matching_route],
                )
            elif "static_files" in handler_dict:
                static_file_path = _build_file_path(
                    request_path,
                    matching_url,
                    static_files=handler_dict["static_files"],
                    working_directory=self._working_directory_mappings[matching_route],
                )

            if static_file_path is not None:
                if os.path.isfile(static_file_path) is True:
                    return _serve_static_file(environ, start_response, static_file_path)

                return _not_found(environ, start_response)

        # Proxy the request
        new_url = _build_proxy_url(
            _build_request_url(environ), self._port_mappings[matching_route]
        )
        return _proxy(environ, start_response, new_url)


def _match_route(to_check, route_rule):
    return fnmatch.fnmatch(to_check, route_rule)


def _match_url(to_check, url_rule):
    path_regex = re.compile(url_rule, re.I)

    if path_regex.match(to_check) is None:
        return False

    return True


def _proxy(environ, start_response, path):
    # Convert the environ into headers
    request_headers = {}

    for environ_variable in environ:
        if environ_variable.startswith("HTTP_") is True:
            request_headers[
                environ_variable[len("HTTP_") :].replace("_", "-")
            ] = environ[environ_variable]

    request = urllib.request.Request(
        path,
        data=environ.get("wsgi.input"),
        headers=request_headers,
        method=environ.get("REQUEST_METHOD"),
    )

    response = urllib.request.urlopen(request)

    start_response(str(response.status) + " " + response.reason, response.getheaders())

    return iter([response.read()])


def _serve_static_file(environ, start_response, file_path):
    def file_wrapper(file_path, block_size=1024):
        # Makes a path into an iterable
        file_object = open(file_path, "rb")
        file_mmap = mmap.mmap(file_object.fileno(), 0, prot=mmap.PROT_READ)

        try:
            data = file_mmap.read(block_size)

            while data:
                yield data
                data = file_mmap.read(block_size)
        finally:
            file_mmap.close()
            file_object.close()

    static_file_mimetype = mimetypes.guess_type(file_path)

    headers = []

    if static_file_mimetype[0] is not None:
        headers.append(("Content-Type", static_file_mimetype[0]))

    if static_file_mimetype[1] is not None:
        headers.append(("Content-Encoding", static_file_mimetype[1]))

    start_response("200 OK", headers)

    if "wsgi.file_wrapper" in environ:
        return environ["wsgi.file_wrapper"](open(file_path, "rb"), 1024)

    # Use our crude memory map iterator instead
    return file_wrapper(file_path)


def _not_found(environ, start_response):
    data = b"Not Found"

    start_response(
        "404 Not Found",
        [("Content-Type", "text/plain"), ("Content-Length", str(len(data)))],
    )

    return iter([data])


def _build_request_url(environ):
    # https://www.python.org/dev/peps/pep-3333/#url-reconstruction
    url = environ["wsgi.url_scheme"] + "://"

    if environ.get("HTTP_HOST"):
        url += environ["HTTP_HOST"]
    else:
        url += environ["SERVER_NAME"]

        if environ["wsgi.url_scheme"] == "https":
            if environ["SERVER_PORT"] != "443":
                url += ":" + environ["SERVER_PORT"]
        else:
            if environ["SERVER_PORT"] != "80":
                url += ":" + environ["SERVER_PORT"]

    url += urllib.parse.quote(environ.get("SCRIPT_NAME", ""))
    url += urllib.parse.quote(environ.get("PATH_INFO", ""))

    if environ.get("QUERY_STRING"):
        url += "?" + environ["QUERY_STRING"]

    return url


def _build_proxy_url(original_url, redirect_port):
    parsed_url = urllib.parse.urlparse(original_url)

    # Replace the original port with the redirect port
    return parsed_url._replace(
        netloc=parsed_url.hostname + ":" + str(redirect_port)
    ).geturl()


def _build_file_path(
    request_path,
    handler_url,
    static_dir=None,
    static_files=None,
    working_directory=None,
):
    if static_dir is not None:
        # Replace the handler_url of the path with static_dir, adding working dir if necessary
        static_file_path = request_path.replace(handler_url, static_dir)
    elif static_files is not None:
        # Map the handler_url regular expression to static_files
        static_file_path = re.sub(handler_url, static_files, request_path)

    else:
        raise ValueError("static_dir or static_files argument required.")

    if working_directory is None:
        return static_file_path

    return os.path.join(working_directory, static_file_path)



# -*- coding: utf-8 -*-

# Copyright (c) 2019, Brandon Nielsen
#
# This file is part of Railyard.
#
# Railyard is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Railyard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Railyard.  If not, see <https://www.gnu.org/licenses/>.

import mimetypes
import os
import unittest
import urllib.request
from unittest.mock import ANY, Mock, patch

from railyard.app import (
    RailyardApplication,
    _build_file_path,
    _build_proxy_url,
    _build_request_url,
    _match_route,
    _match_url,
    _not_found,
    _proxy,
    _serve_static_file,
)


class TestRailyardApplication(unittest.TestCase):
    def test_init(self):
        test_routes = {"*/route1/*": "service1", "*/route2/*": "service2"}
        test_services = {
            "service1": {},
            "service2": {
                "handlers": [
                    {"url": "/service2/static", "static_dir": "static/"},
                    {"url": "/service2/app", "script": "auto"},
                ]
            },
        }
        test_working_directories = {"service1": ".", "service2": "service2/"}
        test_ports = {"service1": 4000, "service2": 5000}

        result_application = RailyardApplication(
            test_routes, test_services, test_working_directories, test_ports
        )

        self.assertEqual(
            result_application._port_mappings, {"*/route1/*": 4000, "*/route2/*": 5000}
        )
        self.assertEqual(
            result_application._url_handlers,
            {
                "*/route1/*": {},
                "*/route2/*": {
                    "/service2/static": {
                        "url": "/service2/static",
                        "static_dir": "static/",
                    },
                    "/service2/app": {"url": "/service2/app", "script": "auto"},
                },
            },
        )
        self.assertEqual(
            result_application._working_directory_mappings,
            {"*/route1/*": ".", "*/route2/*": "service2/"},
        )

    def test_load_config(self):
        result_application = RailyardApplication(
            {}, {}, {}, {}, options={"reload": True, "loglevel": "debug"}
        )

        self.assertEqual(result_application.cfg.reload, True)
        self.assertEqual(result_application.cfg.loglevel, "debug")

    def test_load(self):
        result_application = RailyardApplication({}, {}, {}, {})

        self.assertEqual(result_application.load(), result_application)

    def test_call(self):
        # Test with static_dir
        test_routes = {"*/service/*": "service"}
        test_services = {
            "service": {"handlers": [{"url": "/service/static", "static_dir": "./"}]}
        }
        test_working_directories = {"service": os.path.dirname(__file__)}
        test_ports = {"service": 4000}

        test_application = RailyardApplication(
            test_routes, test_services, test_working_directories, test_ports
        )

        # Bad route
        response_body = test_application.__call__(
            {
                "wsgi.url_scheme": "http",
                "HTTP_HOST": "localhost",
                "PATH_INFO": "/badroute",
            },
            Mock(),
        )

        self.assertEqual(list(response_body), [b"Not Found"])

        # Good route, file exists
        response_body = test_application.__call__(
            {
                "wsgi.url_scheme": "http",
                "HTTP_HOST": "localhost",
                "PATH_INFO": "/service/static/test_static_file",
            },
            Mock(),
        )

        self.assertEqual(list(response_body), [b"test static file content\n"])

        # Good route, file doesn't exist
        response_body = test_application.__call__(
            {
                "wsgi.url_scheme": "http",
                "HTTP_HOST": "localhost",
                "PATH_INFO": "/service/static/bad_file",
            },
            Mock(),
        )

        self.assertEqual(list(response_body), [b"Not Found"])

        # Test with static_files
        test_routes = {"*/service/*": "service"}
        test_services = {
            "service": {
                "handlers": [{"url": "/service/(.*)$", "static_files": "./\\1"}]
            }
        }
        test_working_directories = {"service": os.path.dirname(__file__)}
        test_ports = {"service": 4000}

        test_application = RailyardApplication(
            test_routes, test_services, test_working_directories, test_ports
        )

        # Good route, file exists
        response_body = test_application.__call__(
            {
                "wsgi.url_scheme": "http",
                "HTTP_HOST": "localhost",
                "PATH_INFO": "/service/test_static_file",
            },
            Mock(),
        )

        self.assertEqual(list(response_body), [b"test static file content\n"])

        # Good route, file doesn't exist
        response_body = test_application.__call__(
            {
                "wsgi.url_scheme": "http",
                "HTTP_HOST": "localhost",
                "PATH_INFO": "/service/bad_file",
            },
            Mock(),
        )

        self.assertEqual(list(response_body), [b"Not Found"])

        # Test a proxied response, multiple handler URLs for coverage reasons
        test_routes = {"*/service/*": "service"}
        test_services = {
            "service": {
                "handlers": [
                    {"url": "/images", "static_dir": "static/images"},
                    {"url": "/pdfs", "static_dir": "static/pdfs"},
                ]
            }
        }
        test_working_directories = {"service": "./"}
        test_ports = {"service": 4000}

        test_application = RailyardApplication(
            test_routes, test_services, test_working_directories, test_ports
        )

        mock_url_open_response = Mock()
        mock_url_open_response.status = 200
        mock_url_open_response.reason = "OK"
        mock_url_open_response.getheaders.return_value = {"Proxy": "Headers"}
        mock_url_open_response.read.return_value = b"mock proxied content"

        with patch.object(urllib.request, "Request") as mock_request:
            with patch.object(urllib.request, "urlopen") as mock_url_open:
                mock_request.return_value = Mock()
                mock_url_open.return_value = mock_url_open_response

                response_body = test_application.__call__(
                    {
                        "wsgi.url_scheme": "http",
                        "HTTP_HOST": "localhost",
                        "PATH_INFO": "/service/test",
                    },
                    Mock(),
                )

        self.assertEqual(list(response_body), [b"mock proxied content"])

        # Test a proxied URL
        test_routes = {"*/service/*": "service"}
        test_services = {"service": {"handlers": [{"url": "/.*", "script": "auto"},]}}
        test_working_directories = {"service": "./"}
        test_ports = {"service": 4000}

        test_application = RailyardApplication(
            test_routes, test_services, test_working_directories, test_ports
        )

        mock_url_open_response = Mock()
        mock_url_open_response.status = 200
        mock_url_open_response.reason = "OK"
        mock_url_open_response.getheaders.return_value = {"Proxy": "Headers"}
        mock_url_open_response.read.return_value = b"mock proxied content"

        with patch.object(urllib.request, "Request") as mock_request:
            with patch.object(urllib.request, "urlopen") as mock_url_open:
                mock_request.return_value = Mock()
                mock_url_open.return_value = mock_url_open_response

                response_body = test_application.__call__(
                    {
                        "wsgi.url_scheme": "http",
                        "HTTP_HOST": "localhost",
                        "PATH_INFO": "/service/test",
                    },
                    Mock(),
                )

        self.assertEqual(list(response_body), [b"mock proxied content"])


class TestAppFunctions(unittest.TestCase):
    def test_match_route(self):
        self.assertTrue(_match_route("http://www.test.com/mobile/", "*/mobile/*"))
        self.assertTrue(
            _match_route("http://www.test.com/mobile/index.html", "*/mobile/*")
        )
        self.assertTrue(
            _match_route("http://www.test.com/mobile/sub/index.html", "*/mobile/*")
        )
        self.assertTrue(_match_route("https://www.test.com/mobile/", "*/mobile/*"))
        self.assertTrue(
            _match_route("https://www.test.com/mobile/index.html", "*/mobile/*")
        )
        self.assertTrue(
            _match_route("https://www.test.com/mobile/sub/index.html", "*/mobile/*")
        )
        self.assertTrue(_match_route("http://localhost:8000/mobile/", "*/mobile/*"))
        self.assertTrue(
            _match_route("http://localhost:8000/mobile/index.html", "*/mobile/*")
        )
        self.assertTrue(
            _match_route("http://localhost:8000/mobile/sub/index.html", "*/mobile/*")
        )

        self.assertFalse(_match_route("http://www.test.com/mobile/", "*/work/*"))
        self.assertFalse(
            _match_route("http://www.test.com/mobile/index.html", "*/work/*")
        )
        self.assertFalse(
            _match_route("http://www.test.com/mobile/sub/index.html", "*/work/*")
        )
        self.assertFalse(_match_route("https://www.test.com/mobile/", "*/work/*"))
        self.assertFalse(
            _match_route("https://www.test.com/mobile/index.html", "*/work/*")
        )
        self.assertFalse(
            _match_route("https://www.test.com/mobile/sub/index.html", "*/work/*")
        )
        self.assertFalse(_match_route("http://localhost:8000/mobile/", "*/work/*"))
        self.assertFalse(
            _match_route("http://localhost:8000/mobile/index.html", "*/work/*")
        )
        self.assertFalse(
            _match_route("http://localhost:8000/mobile/sub/index.html", "*/work/*")
        )

    def test_match_url(self):
        self.assertTrue(_match_url("/app/static", "/app/static"))
        self.assertTrue(_match_url("/app/static/test.ext", "/app/static"))
        self.assertTrue(_match_url("/app/static/test/test.ext", "/app/static"))

        self.assertFalse(_match_url("/app/path", "/app/static"))
        self.assertFalse(_match_url("/app/path/test.ext", "/app/static"))
        self.assertFalse(_match_url("/app/path/test/test.ext", "/app/static"))

        self.assertTrue(_match_url("/profile/edit/manager", "/profile/(.*)/(.*)"))
        self.assertFalse(_match_url("/profile/edit", "/profile/(.*)/(.*)"))

    def test_proxy(self):
        mock_start_response = Mock()

        mock_request_object = Mock()

        mock_url_open_response = Mock()
        mock_url_open_response.status = 200
        mock_url_open_response.reason = "OK"
        mock_url_open_response.getheaders.return_value = {"Proxy": "Headers"}
        mock_url_open_response.read.return_value = b"mock proxied content"

        with patch.object(urllib.request, "Request") as mock_request:
            with patch.object(urllib.request, "urlopen") as mock_url_open:
                mock_request.return_value = mock_request_object
                mock_url_open.return_value = mock_url_open_response

                proxied_response_body = _proxy(
                    {
                        "wsgi.input": "test wsgi input",
                        "REQUEST_METHOD": "GET",
                        "HTTP_test_header": "test http header",
                    },
                    mock_start_response,
                    "/test/path",
                )

        mock_request.assert_called_once_with(
            "/test/path",
            data="test wsgi input",
            headers={"test-header": "test http header"},
            method="GET",
        )
        mock_url_open.assert_called_once_with(mock_request_object)
        mock_start_response.assert_called_once_with("200 OK", {"Proxy": "Headers"})

        self.assertEqual(list(proxied_response_body), [b"mock proxied content"])

    def test_serve_static_file(self):
        def test_file_wrapper(file_object, block_size):
            file_content = file_object.read()

            file_object.close()

            return [file_content]

        test_file_path = os.path.join(os.path.dirname(__file__), "test_static_file")

        mock_start_response = Mock()

        with patch.object(mimetypes, "guess_type") as mock_guess_type:
            mock_guess_type.return_value = (None, None)

            response_content = _serve_static_file(
                {}, mock_start_response, test_file_path
            )

        mock_guess_type.assert_called_once_with(test_file_path)
        mock_start_response.assert_called_once_with("200 OK", [])

        self.assertEqual(list(response_content), [b"test static file content\n"])

        # Try again with a mimetype
        mock_start_response = Mock()

        with patch.object(mimetypes, "guess_type") as mock_guess_type:
            mock_guess_type.return_value = ("text/plain", "deflate, gzip")

            response_content = _serve_static_file(
                {}, mock_start_response, test_file_path
            )

        mock_guess_type.assert_called_once_with(test_file_path)
        mock_start_response.assert_called_once_with(
            "200 OK",
            [("Content-Type", "text/plain"), ("Content-Encoding", "deflate, gzip")],
        )

        self.assertEqual(list(response_content), [b"test static file content\n"])

        # Try again with wsgi.file_wrapper
        mock_start_response = Mock()
        mock_file_wrapper = Mock(wraps=test_file_wrapper)

        with patch.object(mimetypes, "guess_type") as mock_guess_type:
            mock_guess_type.return_value = ("text/plain", "deflate, gzip")

            response_content = _serve_static_file(
                {"wsgi.file_wrapper": mock_file_wrapper},
                mock_start_response,
                test_file_path,
            )

        mock_guess_type.assert_called_once_with(test_file_path)
        mock_start_response.assert_called_once_with(
            "200 OK",
            [("Content-Type", "text/plain"), ("Content-Encoding", "deflate, gzip")],
        )
        mock_file_wrapper.assert_called_once_with(ANY, 1024)

        self.assertEqual(list(response_content), [b"test static file content\n"])

    def test_not_found(self):
        mock_start_response = Mock()

        response_body = _not_found(None, mock_start_response)

        mock_start_response.assert_called_once_with(
            "404 Not Found", [("Content-Type", "text/plain"), ("Content-Length", "9")]
        )

        self.assertEqual(list(response_body), [b"Not Found"])

    def test_build_request_url(self):
        self.assertEqual(
            _build_request_url({"wsgi.url_scheme": "http", "HTTP_HOST": "localhost"}),
            "http://localhost",
        )
        self.assertEqual(
            _build_request_url(
                {
                    "wsgi.url_scheme": "https",
                    "SERVER_NAME": "test.com",
                    "SERVER_PORT": "443",
                }
            ),
            "https://test.com",
        )
        self.assertEqual(
            _build_request_url(
                {
                    "wsgi.url_scheme": "https",
                    "SERVER_NAME": "test.com",
                    "SERVER_PORT": "4430",
                }
            ),
            "https://test.com:4430",
        )
        self.assertEqual(
            _build_request_url(
                {
                    "wsgi.url_scheme": "http",
                    "SERVER_NAME": "test.com",
                    "SERVER_PORT": "80",
                }
            ),
            "http://test.com",
        )
        self.assertEqual(
            _build_request_url(
                {
                    "wsgi.url_scheme": "http",
                    "SERVER_NAME": "test.com",
                    "SERVER_PORT": "8000",
                }
            ),
            "http://test.com:8000",
        )
        self.assertEqual(
            _build_request_url(
                {
                    "wsgi.url_scheme": "http",
                    "HTTP_HOST": "localhost",
                    "SCRIPT_NAME": "/script",
                }
            ),
            "http://localhost/script",
        )
        self.assertEqual(
            _build_request_url(
                {
                    "wsgi.url_scheme": "http",
                    "HTTP_HOST": "localhost",
                    "PATH_INFO": "/path_info",
                }
            ),
            "http://localhost/path_info",
        )
        self.assertEqual(
            _build_request_url(
                {
                    "wsgi.url_scheme": "http",
                    "HTTP_HOST": "localhost",
                    "QUERY_STRING": "query1=value1&query2=value2",
                }
            ),
            "http://localhost?query1=value1&query2=value2",
        )

    def test_build_proxy_url(self):
        self.assertEqual(
            _build_proxy_url("http://www.test.com/", 8080), "http://www.test.com:8080/"
        )
        self.assertEqual(
            _build_proxy_url("http://localhost:80/", 8080), "http://localhost:8080/"
        )
        self.assertEqual(
            _build_proxy_url("http://www.test.com/app/path", 8080),
            "http://www.test.com:8080/app/path",
        )
        self.assertEqual(
            _build_proxy_url("http://localhost:80/app/path", 8080),
            "http://localhost:8080/app/path",
        )

    def test_build_file_path(self):
        self.assertEqual(
            _build_file_path(
                "/stylesheets/style.css", "/stylesheets", static_dir="stylesheets"
            ),
            "stylesheets/style.css",
        )
        self.assertEqual(
            _build_file_path(
                "/stylesheets/style.css", "/stylesheets", static_dir="dist/css"
            ),
            "dist/css/style.css",
        )
        self.assertEqual(
            _build_file_path(
                "/stylesheets/style.css",
                "/stylesheets",
                static_dir="stylesheets",
                working_directory="app",
            ),
            "app/stylesheets/style.css",
        )
        self.assertEqual(
            _build_file_path(
                "/stylesheets/style.css",
                "/stylesheets",
                static_dir="dist/css",
                working_directory="app",
            ),
            "app/dist/css/style.css",
        )

        self.assertEqual(
            _build_file_path(
                "/test.gif", r"/(.*\.(gif|png|jpg))$", static_files=r"static/\1"
            ),
            "static/test.gif",
        )
        self.assertEqual(
            _build_file_path(
                "/test.png", r"/(.*\.(gif|png|jpg))$", static_files=r"static/\1"
            ),
            "static/test.png",
        )
        self.assertEqual(
            _build_file_path(
                "/test.jpg", r"/(.*\.(gif|png|jpg))$", static_files=r"static/\1"
            ),
            "static/test.jpg",
        )
        self.assertEqual(
            _build_file_path(
                "/test/test.gif", r"/(.*\.(gif|png|jpg))$", static_files=r"static/\1"
            ),
            "static/test/test.gif",
        )
        self.assertEqual(
            _build_file_path(
                "/test/test.png", r"/(.*\.(gif|png|jpg))$", static_files=r"static/\1"
            ),
            "static/test/test.png",
        )
        self.assertEqual(
            _build_file_path(
                "/test/test.jpg", r"/(.*\.(gif|png|jpg))$", static_files=r"static/\1"
            ),
            "static/test/test.jpg",
        )

        self.assertEqual(
            _build_file_path(
                "/test.gif",
                r"/(.*\.(gif|png|jpg))$",
                static_files=r"static/\1",
                working_directory="app",
            ),
            "app/static/test.gif",
        )
        self.assertEqual(
            _build_file_path(
                "/test.png",
                r"/(.*\.(gif|png|jpg))$",
                static_files=r"static/\1",
                working_directory="app",
            ),
            "app/static/test.png",
        )
        self.assertEqual(
            _build_file_path(
                "/test.jpg",
                r"/(.*\.(gif|png|jpg))$",
                static_files=r"static/\1",
                working_directory="app",
            ),
            "app/static/test.jpg",
        )
        self.assertEqual(
            _build_file_path(
                "/test/test.gif",
                r"/(.*\.(gif|png|jpg))$",
                static_files=r"static/\1",
                working_directory="app",
            ),
            "app/static/test/test.gif",
        )
        self.assertEqual(
            _build_file_path(
                "/test/test.png",
                r"/(.*\.(gif|png|jpg))$",
                static_files=r"static/\1",
                working_directory="app",
            ),
            "app/static/test/test.png",
        )
        self.assertEqual(
            _build_file_path(
                "/test/test.jpg",
                r"/(.*\.(gif|png|jpg))$",
                static_files=r"static/\1",
                working_directory="app",
            ),
            "app/static/test/test.jpg",
        )

        with self.assertRaises(ValueError):
            _build_file_path("/stylesheets/style.css", "/stylesheets")



# -*- coding: utf-8 -*-

# Copyright (c) 2019, Brandon Nielsen
#
# This file is part of Railyard.
#
# Railyard is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Railyard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Railyard.  If not, see <https://www.gnu.org/licenses/>.

from os import path

from setuptools import find_packages, setup

from railyard import __version__

THIS_DIRECTORY = path.abspath(path.dirname(__file__))
with open(path.join(THIS_DIRECTORY, "README.rst"), encoding="utf-8") as f:
    README_TEXT = f.read()

setup(
    name="railyard",
    version=__version__,
    description="Simple local application development server",
    long_description=README_TEXT,
    long_description_content_type="text/x-rst",
    author="Brandon Nielsen",
    author_email="nielsenb@jetfuse.net",
    url="https://bitbucket.org/nielsenb/railyard",
    entry_points={"console_scripts": ["railyard = railyard:main",]},
    python_requires=">=3.7",
    install_requires=["gunicorn", "pyyaml"],
    packages=find_packages(),
    include_package_data=True,
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords="webapp dispatch routing",
    project_urls={
        "Source": "https://bitbucket.org/nielsenb/railyard",
        "Tracker": "https://bitbucket.org/nielsenb/railyard/issues",
    },
)
